#include "chan_misdn_config.h"

#include <asterisk/config.h>
#include <asterisk/channel.h>
#include <asterisk/logger.h>
#include <asterisk/lock.h>
#ifndef ASTERISK_STABLE
#include <asterisk/strings.h>
#endif
#include <stdlib.h>
#include <string.h>

#ifdef ASTERISK_STABLE
#define AST_LOAD_CFG ast_load
#define AST_DESTROY_CFG ast_destroy
#else /* ASTERISK_STABLE */
#include <asterisk/utils.h>
#define AST_LOAD_CFG ast_config_load
#define AST_DESTROY_CFG ast_config_destroy
#endif /* ASTERISK_STABLE */


struct msn_list {
	char *msn;
	struct msn_list *next;
};

struct port_config {
	char *name;
	int *rxgain;
	int *txgain;
	int *te_choose_channel;
	char *context;
	char *language;
	char *callerid;
	int *dialplan; 
	char *natprefix;
	char *internatprefix;
	int *pres;
	int *always_immediate;
	int *immediate;
	int *hold_allowed;
	int *early_bconnect;
	int *use_callingpres;
	int *echocancel;
	int *echocancelwhenbridged;
	int *echotraining;
	struct msn_list *msn_list;
#ifndef ASTERISK_STABLE
	ast_group_t *callgroup;		/* Call group */
	ast_group_t *pickupgroup;	/* Pickup group */
#endif
};

struct general_config {
	int *debug;
	char *tracefile;
	int *trace_calls;
	char *trace_dir;
	int *bridging;
	int *stop_tone_after_first_digit;
	int *append_digits2exten;
	int *l1_info_ok;
	int *clear_l3;
	int *dynamic_crypt;
	char *crypt_prefix;
	char *crypt_keys;
};

/* array of port configs, default is at position 0. */
static struct port_config **port_cfg;
/* max number of available ports, is set on init */
static int max_ports;
/* general config */
static struct general_config *general_cfg;
/* storing the ptp flag separated to save memory */
static int *ptp;

static ast_mutex_t config_mutex; 


static inline void misdn_cfg_lock () {
	ast_mutex_lock(&config_mutex);
}

static inline void misdn_cfg_unlock () {
	ast_mutex_unlock(&config_mutex);
}

static void free_msn_list (struct msn_list* iter) {
	if (iter->next)
		free_msn_list(iter->next);
	if (iter->msn)
		free(iter->msn);
	free(iter);
}

static void free_port_cfg () {

	struct port_config **free_list = (struct port_config **)calloc(max_ports + 1, sizeof(struct port_config *));

	int i, j;

	for (i = 0; i < max_ports; i++) {
		if (port_cfg[i]) {
			for (j = 0; j < max_ports && free_list[j]; j++) {
				if (free_list[j] && free_list[j] == port_cfg[i])
					continue; /* already in list */
				free_list[j] = port_cfg[i];
			}
		}
	}
	
	for (i = 0; i < max_ports; i++) {
		if (free_list[i]) {
			if (free_list[i]->name)
				free(free_list[i]->name);
			if (free_list[i]->rxgain)
				free(free_list[i]->rxgain);
			if (free_list[i]->txgain)
				free(free_list[i]->txgain);
			if (free_list[i]->te_choose_channel)
				free(free_list[i]->te_choose_channel);
			if (free_list[i]->context)
				free(free_list[i]->context);
			if (free_list[i]->language)
				free(free_list[i]->language);
			if (free_list[i]->callerid)
				free(free_list[i]->callerid);
			if (free_list[i]->dialplan)
				free(free_list[i]->dialplan);
			if (free_list[i]->natprefix)
				free(free_list[i]->natprefix);
			if (free_list[i]->internatprefix)
				free(free_list[i]->internatprefix);
			if (free_list[i]->pres)
				free(free_list[i]->pres);
			if (free_list[i]->always_immediate)
				free(free_list[i]->always_immediate);
			if (free_list[i]->immediate)
				free(free_list[i]->immediate);
			if (free_list[i]->hold_allowed)
				free(free_list[i]->hold_allowed);
			if (free_list[i]->early_bconnect)
				free(free_list[i]->early_bconnect);
			if (free_list[i]->use_callingpres)
				free(free_list[i]->use_callingpres);
			if (free_list[i]->echocancel)
				free(free_list[i]->echocancel);
			if (free_list[i]->echocancelwhenbridged)
				free(free_list[i]->echocancelwhenbridged);
			if (free_list[i]->echotraining)
				free(free_list[i]->echotraining);
			if (free_list[i]->msn_list)
				free_msn_list(free_list[i]->msn_list);
#ifndef ASTERISK_STABLE
			if (free_list[i]->callgroup)
				free(free_list[i]->callgroup);
			if (free_list[i]->pickupgroup)
				free(free_list[i]->pickupgroup);
#endif
			free(free_list[i]);
		}
	}
	free(free_list);
}

static void free_general_cfg () {
	if (general_cfg->debug)
		free(general_cfg->debug);
	if (general_cfg->tracefile)
		free(general_cfg->tracefile);
	if (general_cfg->trace_calls)
		free(general_cfg->trace_calls);
	if (general_cfg->trace_dir)
		free(general_cfg->trace_dir);
	if (general_cfg->bridging)
		free(general_cfg->bridging);
	if (general_cfg->stop_tone_after_first_digit)
		free(general_cfg->stop_tone_after_first_digit);
	if (general_cfg->append_digits2exten)
		free(general_cfg->append_digits2exten);
	if (general_cfg->l1_info_ok)
		free(general_cfg->l1_info_ok);
	if (general_cfg->clear_l3)
		free(general_cfg->clear_l3);
	if (general_cfg->dynamic_crypt)
		free(general_cfg->dynamic_crypt);
	if (general_cfg->crypt_prefix)
		free(general_cfg->crypt_prefix);
	if (general_cfg->crypt_keys)
		free(general_cfg->crypt_keys);
}

#define GET_PORTCFG_STRCPY(buf, bufsize, port, item) ({ \
		char *temp; \
		if (port_cfg[port] && port_cfg[port]->item) \
			temp = port_cfg[port]->item; \
		else \
			temp = port_cfg[0]->item; \
		if (!temp || !memccpy((void *)buf, (void *)temp, '\0', bufsize)) \
			memset(buf, 0, 1); \
	})

#define GET_GENCFG_STRCPY(buf, bufsize, item) ({ \
		if (general_cfg && general_cfg->item) { \
			if (!memccpy((void *)buf, (void *)general_cfg->item, '\0', bufsize)) \
				memset(buf, 0, 1); \
		} else \
			memset(buf, 0, 1); \
	})

#define GET_PORTCFG_MEMCPY(buf, bufsize, port, item) ({ \
		typeof(port_cfg[0]->item) temp; \
		if (port_cfg[port] && port_cfg[port]->item) \
			temp = port_cfg[port]->item; \
		else \
			temp = port_cfg[0]->item; \
		if (temp) { \
			int l = sizeof(*temp); \
			if (l > bufsize) \
				memset(buf, 0, bufsize); \
			else \
				memcpy(buf, temp, l); \
		} else \
			memset(buf, 0, bufsize); \
	})

#define GET_GENCFG_MEMCPY(buf, bufsize, item) ({ \
		if (general_cfg && general_cfg->item) { \
			typeof(general_cfg->item) temp = general_cfg->item; \
			int l = sizeof(*temp); \
			if (l > bufsize) \
				memset(buf, 0, bufsize); \
			else \
				memcpy(buf, temp, l); \
		} else \
			memset(buf, 0, bufsize); \
	})

void misdn_cfg_get(int port, enum misdn_cfg_elements elem, void *buf, int bufsize) {
	
	if (!(elem > MISDN_GEN_FIRST) && !misdn_cfg_is_port_valid(port)) {
		memset(buf, 0, bufsize);
		return;
	}

	misdn_cfg_lock();
	
	switch (elem) {
		
		/* port config elements */
			
		case MISDN_CFG_PTP:			if (sizeof(ptp[port]) <= bufsize)
										memcpy(buf, &ptp[port], sizeof(ptp[port]));
									else
										buf = 0; /* error, should not happen */
									break;
		case MISDN_CFG_NAME:		GET_PORTCFG_STRCPY(buf, bufsize, port, name);
									break;
		case MISDN_CFG_RXGAIN:		GET_PORTCFG_MEMCPY(buf, bufsize, port, rxgain);
									break;
		case MISDN_CFG_TXGAIN:		GET_PORTCFG_MEMCPY(buf, bufsize, port, txgain);
									break;
		case MISDN_CFG_TE_CHOOSE_CHANNEL:
									GET_PORTCFG_MEMCPY(buf, bufsize, port, te_choose_channel);
									break;
		case MISDN_CFG_CONTEXT:		GET_PORTCFG_STRCPY(buf, bufsize, port, context);
									break;
		case MISDN_CFG_LANGUAGE:	GET_PORTCFG_STRCPY(buf, bufsize, port, language);
									break;
		case MISDN_CFG_CALLERID:	GET_PORTCFG_STRCPY(buf, bufsize, port, callerid);
									break;
		case MISDN_CFG_DIALPLAN:	GET_PORTCFG_MEMCPY(buf, bufsize, port, dialplan);
									break;
		case MISDN_CFG_NATPREFIX:	GET_PORTCFG_STRCPY(buf, bufsize, port, natprefix);
									break;
		case MISDN_CFG_INTERNATPREFIX:
									GET_PORTCFG_STRCPY(buf, bufsize, port, internatprefix);
									break;
		case MISDN_CFG_PRES:		GET_PORTCFG_MEMCPY(buf, bufsize, port, pres);
									break;
		case MISDN_CFG_ALWAYS_IMMEDIATE:
									GET_PORTCFG_MEMCPY(buf, bufsize, port, always_immediate);
									break;
		case MISDN_CFG_IMMEDIATE:	GET_PORTCFG_MEMCPY(buf, bufsize, port, immediate);
									break;
		case MISDN_CFG_HOLD_ALLOWED:
									GET_PORTCFG_MEMCPY(buf, bufsize, port, hold_allowed);
									break;
		case MISDN_CFG_EARLY_BCONNECT:
									GET_PORTCFG_MEMCPY(buf, bufsize, port, early_bconnect);
									break;
		case MISDN_CFG_USE_CALLINGPRES:
									GET_PORTCFG_MEMCPY(buf, bufsize, port, use_callingpres);
									break;
		case MISDN_CFG_ECHOCANCEL:
									GET_PORTCFG_MEMCPY(buf, bufsize, port, echocancel );
									break;
		case MISDN_CFG_ECHOCANCELWHENBRIDGED:
									GET_PORTCFG_MEMCPY(buf, bufsize, port, echocancelwhenbridged);
									break;
		case MISDN_CFG_ECHOTRAINING:
									GET_PORTCFG_MEMCPY(buf,  bufsize, port, echotraining);
									break;
#ifndef ASTERISK_STABLE
		case MISDN_CFG_CALLGROUP:	GET_PORTCFG_MEMCPY(buf, bufsize, port, callgroup);
									break;
		case MISDN_CFG_PICKUPGROUP:	GET_PORTCFG_MEMCPY(buf, bufsize, port, pickupgroup);
									break;
#endif
		
		/* general config elements */
			
		case MISDN_GEN_DEBUG:		GET_GENCFG_MEMCPY(buf, bufsize, debug);
									break;
		case MISDN_GEN_TRACEFILE:	GET_GENCFG_STRCPY(buf, bufsize, tracefile);
									break;
		case MISDN_GEN_TRACE_CALLS: GET_GENCFG_MEMCPY(buf, bufsize, trace_calls);
									break;
		case MISDN_GEN_TRACE_DIR:	GET_GENCFG_STRCPY(buf, bufsize, trace_dir);
									break;
		case MISDN_GEN_BRIDGING:	GET_GENCFG_MEMCPY(buf, bufsize, bridging);
									break;
		case MISDN_GEN_STOP_TONE:	GET_GENCFG_MEMCPY(buf, bufsize, stop_tone_after_first_digit);
									break;
		case MISDN_GEN_APPEND_DIGITS2EXTEN: 
									GET_GENCFG_MEMCPY(buf, bufsize, append_digits2exten);
									break;
		case MISDN_GEN_L1_INFO_OK:  GET_GENCFG_MEMCPY(buf, bufsize, l1_info_ok);
									break;
		case MISDN_GEN_CLEAR_L3:	GET_GENCFG_MEMCPY(buf, bufsize, clear_l3);
									break;
		case MISDN_GEN_DYNAMIC_CRYPT:
									GET_GENCFG_MEMCPY(buf, bufsize, dynamic_crypt);
									break;
		case MISDN_GEN_CRYPT_PREFIX:
									GET_GENCFG_STRCPY(buf, bufsize, crypt_prefix);
									break;
		case MISDN_GEN_CRYPT_KEYS:  GET_GENCFG_STRCPY(buf, bufsize, crypt_keys);
									break;
		default:					memset(buf, 0, bufsize);
	}
	
	misdn_cfg_unlock();
}

int misdn_cfg_is_msn_valid (int port, char* msn) {
	
	if (!misdn_cfg_is_port_valid(port))
		return 0;
	
	struct msn_list *iter;
	
	misdn_cfg_lock();
	
	if (port_cfg[port]->msn_list)
		iter = port_cfg[port]->msn_list;
	else
		iter = port_cfg[0]->msn_list;
	for (; iter; iter = iter->next) 
		if (*(iter->msn) == '*' || !strcasecmp(iter->msn, msn)) {
			misdn_cfg_unlock();
			return 1;
		}
	
	misdn_cfg_unlock();
	
	return 0;
}

int misdn_cfg_is_port_valid (int port) {
	
	misdn_cfg_lock();
	
	if (port < 1 || port > max_ports) {
		misdn_cfg_unlock();
		return 0;
	}

	int valid = (port_cfg[port] != NULL);

	misdn_cfg_unlock();

	return valid;
}

void misdn_cfg_get_ports_string (char *ports) {
	*ports = 0;
	char tmp[16];
	
	misdn_cfg_lock();

	int i = 1;
	for (; i <= max_ports; i++) {
		if (port_cfg[i]) {
			if (ptp[i])
				sprintf(tmp, "%dptp,", i);
			else
				sprintf(tmp, "%d,", i);
			strcat(ports, tmp);
		}
	}
	
	misdn_cfg_unlock();
	
	ports[strlen(ports)-1] = 0;
}

#define GET_CFG_STRING(typestr, type) ({ \
		if (port_cfg[port] && port_cfg[port]->type) \
			snprintf(buf, bufsize, "%s " #typestr ": %s", begin, port_cfg[port]->type); \
		else \
			snprintf(buf, bufsize, "%s " #typestr ": %s", begin, port_cfg[0]->type); \
	}) \

#define GET_GEN_STRING(typestr, type) ({ \
		snprintf(buf, bufsize, "%s " #typestr ": %s", begin, general_cfg->type ? general_cfg->type : "not set"); \
	}) \

#define GET_CFG_INT(typestr, type) ({ \
		if (port_cfg[port] && port_cfg[port]->type) \
			snprintf(buf, bufsize, "%s " #typestr ": %d", begin, *port_cfg[port]->type); \
		else \
			snprintf(buf, bufsize, "%s " #typestr ": %d", begin, *port_cfg[0]->type); \
	}) \

#define GET_GEN_INT(typestr, type) ({ \
		snprintf(buf, bufsize, "%s " #typestr ": %d", begin, general_cfg->type ? *general_cfg->type : 0); \
	}) \

#define GET_CFG_BOOL(typestr, type, yes, no) ({ \
		int bool; \
		if (port_cfg[port] && port_cfg[port]->type) \
			bool = *port_cfg[port]->type; \
		else \
			bool = *port_cfg[0]->type; \
		snprintf(buf, bufsize, "%s " #typestr ": %s", begin, bool ? #yes : #no); \
	}) \

#define GET_GEN_BOOL(typestr, type, yes, no) ({ \
		snprintf(buf, bufsize, "%s " #typestr ": %s", begin, general_cfg->type ? (*general_cfg->type ? #yes : #no) : "not set"); \
	}) \

#define GET_CFG_AST_GROUP_T(typestr, type) ({ \
		ast_group_t *tmp; \
		if (port_cfg[port] && port_cfg[port]->type) \
			tmp = port_cfg[port]->type; \
		else \
			tmp = port_cfg[0]->type; \
		if (tmp) { \
			char tmpbuf[256]; \
			snprintf(buf, bufsize, "%s " #typestr ": %s", begin, ast_print_group(tmpbuf, sizeof(tmpbuf), *tmp)); \
		} else \
			snprintf(buf, bufsize, "%s " #typestr ": %s", begin, "none"); \
	}) \

void misdn_cfg_get_config_string(int port, enum misdn_cfg_elements elem, char* buf, int bufsize) {

	if (!(elem > MISDN_GEN_FIRST) && !misdn_cfg_is_port_valid(port)) {
		*buf = 0;
		return;
	}
	
	char begin[] = " -> ";
	
	misdn_cfg_lock();
	
	switch (elem) {
		
		case MISDN_CFG_PTP:			snprintf(buf, bufsize, "%s PTP: %s", begin, ptp[port] ? "yes" : "no");
									break;
		case MISDN_CFG_NAME:		GET_CFG_STRING(GROUPNAME, name);
									break;
		case MISDN_CFG_RXGAIN:		GET_CFG_INT(RXGAIN, rxgain);
									break;
		case MISDN_CFG_TXGAIN:		GET_CFG_INT(TXGAIN, txgain);
									break;
		case MISDN_CFG_TE_CHOOSE_CHANNEL:
									GET_CFG_BOOL(TE_CHOOSE_CHANNEL, te_choose_channel, yes, no);
									break;
		case MISDN_CFG_CONTEXT:		GET_CFG_STRING(CONTEXT, context);
									break;
		case MISDN_CFG_LANGUAGE:	GET_CFG_STRING(LANGUAGE, language);
									break;
		case MISDN_CFG_CALLERID:	GET_CFG_STRING(CALLERID, callerid);
									break;
		case MISDN_CFG_DIALPLAN:	GET_CFG_INT(DIALPLAN, dialplan);
									break;
		case MISDN_CFG_NATPREFIX:	GET_CFG_STRING(NATIONALPREFIX, natprefix);
									break;
		case MISDN_CFG_INTERNATPREFIX:
									GET_CFG_STRING(INTERNATIONALPREFIX, internatprefix);
									break;
		case MISDN_CFG_PRES:		GET_CFG_BOOL(PRESENTATION, pres, allowed, not_screened);
									break;
		case MISDN_CFG_ALWAYS_IMMEDIATE:
									GET_CFG_BOOL(ALWAYS_IMMEDIATE, always_immediate, yes, no);
									break;
		case MISDN_CFG_IMMEDIATE:	GET_CFG_BOOL(IMMEDIATE, immediate, yes, no);
									break;
		case MISDN_CFG_HOLD_ALLOWED:
									GET_CFG_BOOL(HOLD_ALLOWED, hold_allowed, yes, no);
									break;
		case MISDN_CFG_EARLY_BCONNECT:
									GET_CFG_BOOL(EARLY_BCONNECT, early_bconnect, yes, no);
									break;
		case MISDN_CFG_USE_CALLINGPRES:
									GET_CFG_BOOL(USE_CALLINGPRES, use_callingpres, yes, no);
									break;
		case MISDN_CFG_ECHOCANCEL:
									GET_CFG_INT(ECHOCANCEL, echocancel );
									break;
		case MISDN_CFG_ECHOCANCELWHENBRIDGED:
									GET_CFG_BOOL(ECHOCANCELWHENBRIDGED, echocancelwhenbridged, yes, no);
									break;
		case MISDN_CFG_ECHOTRAINING:
									GET_CFG_BOOL(ECHOTRAINING, echotraining, yes, no);
									break;
									
#ifndef ASTERISK_STABLE
		case MISDN_CFG_CALLGROUP:	GET_CFG_AST_GROUP_T(CALLINGGROUP, callgroup);
									break;
		case MISDN_CFG_PICKUPGROUP:	GET_CFG_AST_GROUP_T(PICKUPGROUP, pickupgroup);
									break;
#endif

		case MISDN_CFG_MSNS:		{
										char tmpbuffer[BUFFERSIZE];
										tmpbuffer[0] = 0;
										struct msn_list *iter;
										if (port_cfg[port]->msn_list)
											iter = port_cfg[port]->msn_list;
										else
											iter = port_cfg[0]->msn_list;
										if (iter) {
											for (; iter; iter = iter->next)
												sprintf(tmpbuffer, "%s%s, ", tmpbuffer, iter->msn);
											tmpbuffer[strlen(tmpbuffer)-2] = 0;
										}
										snprintf(buf, bufsize, "%s MSNs: %s", begin, *tmpbuffer ? tmpbuffer : "none"); \
									}
									break;
		
		/* general config elements */
		
		case MISDN_GEN_DEBUG:		GET_GEN_INT(DEBUG_LEVEL, debug);
									break;
		case MISDN_GEN_TRACEFILE:	GET_GEN_STRING(TRACEFILE, tracefile);
									break;
		case MISDN_GEN_TRACE_CALLS: GET_GEN_BOOL(TRACE_CALLS, trace_calls, true, false);
									break;
		case MISDN_GEN_TRACE_DIR:	GET_GEN_STRING(TRACE_DIR, trace_dir);
									break;
		case MISDN_GEN_BRIDGING:	GET_GEN_BOOL(BRIDGING, bridging, yes, no);
									break;
		case MISDN_GEN_STOP_TONE:	GET_GEN_BOOL(STOP_TONE_AFTER_FIRST_DIGIT, stop_tone_after_first_digit, yes, no);
									break;
		case MISDN_GEN_APPEND_DIGITS2EXTEN: 
									GET_GEN_BOOL(APPEND_DIGITS2EXTEN, append_digits2exten, yes, no);
									break;
		case MISDN_GEN_L1_INFO_OK:  GET_GEN_BOOL(L1_INFO_OK, l1_info_ok, yes, no);
									break;
		case MISDN_GEN_CLEAR_L3:	GET_GEN_BOOL(CLEAR_L3, clear_l3, yes, no);
									break;
		case MISDN_GEN_DYNAMIC_CRYPT:
									GET_GEN_BOOL(DYNAMIC_CRYPT,dynamic_crypt, yes, no);
									break;
		case MISDN_GEN_CRYPT_PREFIX:
									GET_GEN_STRING(CRYPT_PREFIX, crypt_prefix);
									break;
		case MISDN_GEN_CRYPT_KEYS:  GET_GEN_STRING(CRYPT_KEYS, crypt_keys);
									break;
									
		default:					*buf = 0;
									break;
	}

	misdn_cfg_unlock();
}

int misdn_cfg_get_next_port (int port) {
	
	misdn_cfg_lock();
	
	for (port++; port <= max_ports; port++) {
		if (port_cfg[port]) {
			misdn_cfg_unlock();
			return port;
		}
	}
	
	misdn_cfg_unlock();
	
	return -1;
}

static void build_general_config(struct ast_variable *v) {

	if (!v) 
		return;

	while (v) {

		if (!strcasecmp(v->name, "debug")) {
			general_cfg->debug = (int *)malloc(sizeof(int));
			*(general_cfg->debug) = (int)atoi(v->value);
		}
		if (!strcasecmp(v->name, "tracefile")) {
			int l = strlen(v->value);
			general_cfg->tracefile = (char*)calloc(l+1,sizeof(char));
			strncpy(general_cfg->tracefile, v->value, l);
		}
		if (!strcasecmp(v->name, "trace_calls")) {
			general_cfg->trace_calls = (int *)malloc(sizeof(int));
			if (!strcasecmp(v->value, "true"))
					*(general_cfg->trace_calls) = 1;
			else
					*(general_cfg->trace_calls) = 0;
		}
		if (!strcasecmp(v->name, "trace_dir")) {
			int l = strlen(v->value);
			general_cfg->trace_dir = (char*)calloc(l+1,sizeof(char));
			strncpy(general_cfg->trace_dir, v->value, l);
		}
		if (!strcasecmp(v->name, "bridging")) {
			general_cfg->bridging = (int *)malloc(sizeof(int));
			if (!strcasecmp(v->value, "yes"))
					*(general_cfg->bridging) = 1;
			else
					*(general_cfg->bridging) = 0;
		}
		if (!strcasecmp(v->name, "stop_tone_after_first_digit")) {
			general_cfg->stop_tone_after_first_digit = (int *)malloc(sizeof(int));
			if (!strcasecmp(v->value, "yes"))
					*(general_cfg->stop_tone_after_first_digit) = 1;
			else
					*(general_cfg->stop_tone_after_first_digit) = 0;
		}
		if (!strcasecmp(v->name, "append_digits2exten")) {
			general_cfg->append_digits2exten = (int *)malloc(sizeof(int));
			if (!strcasecmp(v->value, "yes"))
					*(general_cfg->append_digits2exten) = 1;
			else
					*(general_cfg->append_digits2exten) = 0;
		}
		if (!strcasecmp(v->name, "l1_info_ok")) {
			general_cfg->l1_info_ok = (int *)malloc(sizeof(int));
			if (!strcasecmp(v->value, "yes"))
					*(general_cfg->l1_info_ok) = 1;
			else
					*(general_cfg->l1_info_ok) = 0;
		}
		if (!strcasecmp(v->name, "clear_l3")) {
			general_cfg->clear_l3 = (int *)malloc(sizeof(int));
			if (!strcasecmp(v->value, "no"))
					*(general_cfg->clear_l3) = 0;
			else
					*(general_cfg->clear_l3) = 1;
		}
		if (!strcasecmp(v->name, "dynamic_crypt")) {
			general_cfg->dynamic_crypt = (int *)malloc(sizeof(int));
			if (!strcasecmp(v->value, "no"))
					*(general_cfg->dynamic_crypt) = 0;
			else
					*(general_cfg->dynamic_crypt) = 1;
		}
		if (!strcasecmp(v->name, "crypt_prefix")) {
			int l = strlen(v->value);
			general_cfg->crypt_prefix = (char*)calloc(l+1,sizeof(char));
			strncpy(general_cfg->crypt_prefix, v->value, l);
		}
		if (!strcasecmp(v->name, "crypt_keys")) {
			int l = strlen(v->value);
			general_cfg->crypt_keys = (char*)calloc(l+1,sizeof(char));
			strncpy(general_cfg->crypt_keys, v->value, l);
		}
		v  = v->next;
	}
}

static void build_port_config(struct ast_variable *v, char *cat) {
	if (!v || !cat)
		return;

	int cfg_for_ports[max_ports + 1];
	int i = 0;
	for (; i < (max_ports + 1); i++) {
		cfg_for_ports[i] = 0;
	}
	
	/* we store the default config at position 0 */
	if (!strcasecmp(cat, "default")) {
		cfg_for_ports[0] = 1;
	}

	struct port_config* new = (struct port_config *)calloc(1, sizeof(struct port_config));
	
	{
		int l = strlen(cat);
		new->name = (char *)calloc(l+1, sizeof(char));
		strncpy(new->name, cat, l);
	}
	
	while (v) {
		if (!strcasecmp(v->name, "ports")) {
			/* TODO check for value not beeing set, like PORTS= */
			char *iter;
			char *value = v->value;
			while ((iter = strchr(value, ',')) != NULL) {
				*iter = 0;
				/* strip spaces */
				while (*value && *value == ' ') {
					value++;
				}
				/* TODO check for char not 0-9 */

				if (*value){
					int p = atoi(value);
					if (p <= max_ports && p > 0) {
						cfg_for_ports[p] = 1;
						if (strstr(value, "ptp"))
							ptp[p] = 1;
					}
					value = ++iter;
				}
			}

			/* the remaining or the only one */
			
			/* strip spaces */
			while (*value && *value == ' ') {
				value++;
			}
			/* TODO check for char not 0-9 */
			if (*value) {
				int p = atoi(value);
				if (p <= max_ports && p > 0) {
					cfg_for_ports[p] = 1;
					if (strstr(value, "ptp"))
						ptp[p] = 1;
				}
			}
		}
			
		if (!strcasecmp(v->name, "context")) {
			int l = strlen(v->value);
			if (l) {
				new->context = (char *)calloc(l+1, sizeof(char));
				strncpy(new->context,v->value,l);
			}
		}

		if (!strcasecmp(v->name, "dialplan")) {
			new->dialplan = (int *)malloc(sizeof(int));
			*(new->dialplan) = atoi(v->value);
		}
    
		if (!strcasecmp(v->name, "nationalprefix")) {
			int l = strlen(v->value);
			if (l) {
				new->natprefix = (char *)calloc(l+1, sizeof(char));
				strncpy(new->natprefix,v->value,l);
			}
		}
    
		if (!strcasecmp(v->name, "internationalprefix")) {
			int l = strlen(v->value);
			if (l) {
				new->internatprefix = (char *)calloc(l+1, sizeof(char));
				strncpy(new->internatprefix,v->value,l);
			}
		}
    
		if (!strcasecmp(v->name, "language")) {
			int l = strlen(v->value);
			if (l) {
				new->language = (char *)calloc(l+1, sizeof(char));
				strncpy(new->language,v->value,l);
			}
		}

		if (!strcasecmp(v->name, "presentation")) {
			if (v->value && strlen(v->value)) {
				new->pres = (int *)malloc(sizeof(int));
				if (!strcasecmp(v->value, "allowed")) {
					*(new->pres) = 1;
				}
				/* TODO: i assume if it is not "allowed", it is "not_screened" */
				else {
					*(new->pres) = 0;
				}
			}
		}

		if (!strcasecmp(v->name, "rxgain")) {
			new->rxgain = (int *)malloc(sizeof(int));
			*(new->rxgain) = atoi(v->value);
		}
		
		if (!strcasecmp(v->name, "txgain")) {
			new->txgain = (int *)malloc(sizeof(int));
			*(new->txgain) = atoi(v->value);
		}
		
		if (!strcasecmp(v->name, "te_choose_channel")) {
			new->te_choose_channel = (int *)malloc(sizeof(int));
			*(new->te_choose_channel) = ast_true(v->value);
		}
		
		if (!strcasecmp(v->name, "immediate")) {
			new->immediate = (int*)malloc(sizeof(int));
			if (ast_true(v->value))
				*(new->immediate) = 1;
			else
				*(new->immediate) = 0;
		}

		if (!strcasecmp(v->name, "always_immediate")) {
			new->always_immediate = (int*)malloc(sizeof(int));
			if (ast_true(v->value))
				*(new->always_immediate) = 1;
			else
				*(new->always_immediate) = 0;
		}
		
		if (!strcasecmp(v->name, "hold_allowed")) {
			new->hold_allowed = (int*)malloc(sizeof(int));
			if (ast_true(v->value))
				*(new->hold_allowed) = 1;
			else
				*(new->hold_allowed) = 0;
		}
		
		if (!strcasecmp(v->name, "early_bconnect")) {
			new->early_bconnect = (int*)malloc(sizeof(int));
			if (ast_true(v->value))
				*(new->early_bconnect) = 1;
			else
				*(new->early_bconnect) = 0;
		}
		
		if (!strcasecmp(v->name, "use_callingpres")) {
			new->use_callingpres = (int*)malloc(sizeof(int));
			if (ast_true(v->value))
				*(new->use_callingpres) = 1;
			else
				*(new->use_callingpres) = 0;
		}
		
		if (!strcasecmp(v->name, "echocancel")) {
			new->echocancel= (int*)malloc(sizeof(int));
			*(new->echocancel) = atoi(v->value);
		}

		if (!strcasecmp(v->name, "echocancelwhenbridged")) {
			new->echocancelwhenbridged= (int*)malloc(sizeof(int));
			*(new->echocancelwhenbridged) = ast_true(v->value);
		}
		
		if (!strcasecmp(v->name, "echotraining")) {
			new->echotraining= (int*)malloc(sizeof(int));
			*(new->echotraining) = ast_true(v->value);
		}
		
		if (!strcasecmp(v->name, "callerid")) {
			int l = strlen(v->value);
			if (l) {
				new->callerid = (char *)calloc(l+1, sizeof(char));
				strncpy(new->callerid,v->value,l);
			}
		}
    
		if (!strcasecmp(v->name, "msns")) {
			/* TODO check for value not beeing set, like msns= */
			char *iter;
			char *value = v->value;

			while ((iter = strchr(value, ',')) != NULL) {
				*iter = 0;
				/* strip spaces */
				while (*value && *value == ' ') {
					value++;
				}
				/* TODO check for char not 0-9 */

				if (*value){
					int l = strlen(value);
					if (l) {
						struct msn_list *ml = (struct msn_list *)calloc(1, sizeof(struct msn_list));
						ml->msn = (char *)calloc(l+1, sizeof(char));
						strncpy(ml->msn,value,l);
						ml->next = new->msn_list;
						new->msn_list = ml;
					}
					value = ++iter;
				}
			}

			/* the remaining or the only one */
			
			/* strip spaces */
			while (*value && *value == ' ') {
				value++;
			}
			/* TODO check for char not 0-9 */
			if (*value) {
				int l = strlen(value);
				if (l) {
					struct msn_list *ml = (struct msn_list *)calloc(1, sizeof(struct msn_list));
					ml->msn = (char *)calloc(l+1, sizeof(char));
					strncpy(ml->msn,value,l);
					ml->next = new->msn_list;
					new->msn_list = ml;
				}
			}
		}
			
#ifndef ASTERISK_STABLE
		if (!strcasecmp(v->name, "callgroup")) {
			new->callgroup = (ast_group_t *)malloc(sizeof(ast_group_t));
			*(new->callgroup)=ast_get_group(v->value);
		}
		
		if (!strcasecmp(v->name, "pickupgroup")) {
			new->pickupgroup = (ast_group_t *)malloc(sizeof(ast_group_t));
			*(new->pickupgroup)=ast_get_group(v->value);
		}
#endif
    
		v=v->next;
	}

	/* store the new config in our array of port configs */
	for (i = 0; i < (max_ports + 1); i++) {
		if (cfg_for_ports[i])
			port_cfg[i] = new;
	}
}


static void fill_defaults () {
	
	/* general defaults */
	if (!general_cfg->debug)
		general_cfg->debug = (int*)calloc(1, sizeof(int));
	if (!general_cfg->trace_calls)
		general_cfg->trace_calls = (int*)calloc(1, sizeof(int));
	if (!general_cfg->trace_dir) {
		general_cfg->trace_dir = (char *)malloc(10 * sizeof(char));
		sprintf(general_cfg->trace_dir, "/var/log/");
	}
	if (!general_cfg->bridging) {
		general_cfg->bridging = (int*)calloc(1, sizeof(int));
		*general_cfg->bridging = 1;
	}
	if (!general_cfg->stop_tone_after_first_digit) {
		general_cfg->stop_tone_after_first_digit = (int*)calloc(1, sizeof(int));
		*general_cfg->stop_tone_after_first_digit = 1;
	}
	if (!general_cfg->append_digits2exten) {
		general_cfg->append_digits2exten = (int*)calloc(1, sizeof(int));
		*general_cfg->append_digits2exten = 1;
	}
	if (!general_cfg->l1_info_ok)
		general_cfg->l1_info_ok = (int*)calloc(1, sizeof(int));
	if (!general_cfg->clear_l3)
		general_cfg->clear_l3 =(int*)calloc(1, sizeof(int));
	if (!general_cfg->dynamic_crypt)
		general_cfg->dynamic_crypt = (int*)calloc(1, sizeof(int));

	/* defaults for default port config */
	if (!port_cfg[0])
		port_cfg[0] = (struct port_config*)calloc(1, sizeof(struct port_config));
	if (!port_cfg[0]->rxgain)
		port_cfg[0]->rxgain = (int *)calloc(1, sizeof(int));
	if (!port_cfg[0]->txgain)
		port_cfg[0]->txgain = (int *)calloc(1, sizeof(int));
	if (!port_cfg[0]->te_choose_channel)
		port_cfg[0]->te_choose_channel = (int *)calloc(1, sizeof(int));
	if (!port_cfg[0]->context) {
		port_cfg[0]->context = (char *)malloc(8 * sizeof(char));
		sprintf(port_cfg[0]->context, "default");
	}
	if (!port_cfg[0]->language) {
		port_cfg[0]->language = (char *)malloc(3 * sizeof(char));
		sprintf(port_cfg[0]->language, "en");
	}
	if (!port_cfg[0]->callerid) {
		port_cfg[0]->callerid = (char *)calloc(1, sizeof(char));
	}
	if (!port_cfg[0]->dialplan)
		port_cfg[0]->dialplan = (int *)calloc(1, sizeof(int));
	if (!port_cfg[0]->natprefix) {
		port_cfg[0]->natprefix = (char *)malloc(2 * sizeof(char));
		sprintf(port_cfg[0]->natprefix, "0");
	}
	if (!port_cfg[0]->internatprefix) {
		port_cfg[0]->internatprefix = (char *)malloc(3 * sizeof(char));
		sprintf(port_cfg[0]->internatprefix, "00");
	}
	if (!port_cfg[0]->pres) {
		port_cfg[0]->pres = (int *)malloc(sizeof(int));
		*port_cfg[0]->pres = 1;
	}
	if (!port_cfg[0]->always_immediate)
		port_cfg[0]->always_immediate = (int *)calloc(1, sizeof(int));
	if (!port_cfg[0]->immediate)
		port_cfg[0]->immediate = (int *)calloc(1, sizeof(int));
	if (!port_cfg[0]->hold_allowed)
		port_cfg[0]->hold_allowed = (int *)calloc(1, sizeof(int));
	if (!port_cfg[0]->early_bconnect) {
		port_cfg[0]->early_bconnect = (int *)calloc(1, sizeof(int));
		*port_cfg[0]->early_bconnect = 1;
	}
	if (!port_cfg[0]->echocancel)
		port_cfg[0]->echocancel=(int *)calloc(1, sizeof(int));
	if (!port_cfg[0]->echocancelwhenbridged)
		port_cfg[0]->echocancelwhenbridged=(int *)calloc(1, sizeof(int));
	if (!port_cfg[0]->echotraining) {
		port_cfg[0]->echotraining=(int *)calloc(1, sizeof(int));
		*port_cfg[0]->echotraining = 1;
	}
	
	if (!port_cfg[0]->use_callingpres) {
		port_cfg[0]->use_callingpres = (int *)calloc(1, sizeof(int));
		*port_cfg[0]->use_callingpres = 1;
	}
	if (!port_cfg[0]->msn_list) {
		port_cfg[0]->msn_list = (struct msn_list *)malloc(sizeof(struct msn_list));
		port_cfg[0]->msn_list->next = NULL;
		port_cfg[0]->msn_list->msn = (char *)calloc(2, sizeof(char));
		*(port_cfg[0]->msn_list->msn) = '*';
	}
}

void misdn_cfg_reload () {
	misdn_cfg_init (0);
}

void misdn_cfg_destroy () {

	misdn_cfg_lock();
	
	free_port_cfg();
	free_general_cfg();
	
	free(port_cfg);
	free(general_cfg);
	free(ptp);
	
	misdn_cfg_unlock();
}

void misdn_cfg_init (int this_max_ports)
{
	char config[]="misdn.conf";
	
	struct ast_config *cfg;
	cfg = AST_LOAD_CFG(config);
	if (!cfg) {
		ast_log(LOG_WARNING,"no misdn.conf ?\n");
		return;
	}

	misdn_cfg_lock();
	
	if (this_max_ports) {
		/* this is the first run */
		max_ports = this_max_ports;
		port_cfg = (struct port_config **)calloc(max_ports + 1, sizeof(struct port_config *));
		general_cfg = (struct general_config*)calloc(1, sizeof(struct general_config));
		ptp = (int *)calloc(max_ports + 1, sizeof(int));
	}
	else {
		free_port_cfg();
		free_general_cfg();
		port_cfg = memset(port_cfg, 0, sizeof(struct port_config *) * (max_ports + 1));
		general_cfg = memset(general_cfg, 0, sizeof(struct general_config));
		ptp = memset(ptp, 0, sizeof(int) * (max_ports + 1));
	}
	
	char *cat;
	cat = ast_category_browse(cfg, NULL);

	while(cat) {
		struct ast_variable *v=ast_variable_browse(cfg,cat);
		if (!strcasecmp(cat,"general")) {
			build_general_config (v);
		} else {
			build_port_config (v, cat);
		}
		cat=ast_category_browse(cfg,cat);
	}

	fill_defaults();
	
	misdn_cfg_unlock();
	
	AST_DESTROY_CFG(cfg);
}
